# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: 0.43.09-v201909272304
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat
from six import iteritems
import re


class AnnotationInputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'backup_name': 'str',
        'created_by_id': 'str',
        'description': 'str',
        'discoverable': 'bool',
        'document': 'str',
        'interests': 'list[AnnotationInterestInputV1]',
        'name': 'str',
        'replies_to': 'str',
        'type': 'str'
    }

    attribute_map = {
        'backup_name': 'backupName',
        'created_by_id': 'createdById',
        'description': 'description',
        'discoverable': 'discoverable',
        'document': 'document',
        'interests': 'interests',
        'name': 'name',
        'replies_to': 'repliesTo',
        'type': 'type'
    }

    def __init__(self, backup_name=None, created_by_id=None, description=None, discoverable=None, document=None, interests=None, name=None, replies_to=None, type=None):
        """
        AnnotationInputV1 - a model defined in Swagger
        """

        self._backup_name = None
        self._created_by_id = None
        self._description = None
        self._discoverable = None
        self._document = None
        self._interests = None
        self._name = None
        self._replies_to = None
        self._type = None

        if backup_name is not None:
          self.backup_name = backup_name
        if created_by_id is not None:
          self.created_by_id = created_by_id
        if description is not None:
          self.description = description
        if discoverable is not None:
          self.discoverable = discoverable
        if document is not None:
          self.document = document
        if interests is not None:
          self.interests = interests
        if name is not None:
          self.name = name
        if replies_to is not None:
          self.replies_to = replies_to
        if type is not None:
          self.type = type

    @property
    def backup_name(self):
        """
        Gets the backup_name of this AnnotationInputV1.
        The name of a Document backup to restore, when updating an annotation.

        :return: The backup_name of this AnnotationInputV1.
        :rtype: str
        """
        return self._backup_name

    @backup_name.setter
    def backup_name(self, backup_name):
        """
        Sets the backup_name of this AnnotationInputV1.
        The name of a Document backup to restore, when updating an annotation.

        :param backup_name: The backup_name of this AnnotationInputV1.
        :type: str
        """

        self._backup_name = backup_name

    @property
    def created_by_id(self):
        """
        Gets the created_by_id of this AnnotationInputV1.
        The ID of the User that created this annotation.

        :return: The created_by_id of this AnnotationInputV1.
        :rtype: str
        """
        return self._created_by_id

    @created_by_id.setter
    def created_by_id(self, created_by_id):
        """
        Sets the created_by_id of this AnnotationInputV1.
        The ID of the User that created this annotation.

        :param created_by_id: The created_by_id of this AnnotationInputV1.
        :type: str
        """

        self._created_by_id = created_by_id

    @property
    def description(self):
        """
        Gets the description of this AnnotationInputV1.
        Clarifying information or other plain language description of this asset. An input of just whitespace is equivalent to a null input.

        :return: The description of this AnnotationInputV1.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this AnnotationInputV1.
        Clarifying information or other plain language description of this asset. An input of just whitespace is equivalent to a null input.

        :param description: The description of this AnnotationInputV1.
        :type: str
        """

        self._description = description

    @property
    def discoverable(self):
        """
        Gets the discoverable of this AnnotationInputV1.
        Whether this annotation is discoverable.

        :return: The discoverable of this AnnotationInputV1.
        :rtype: bool
        """
        return self._discoverable

    @discoverable.setter
    def discoverable(self, discoverable):
        """
        Sets the discoverable of this AnnotationInputV1.
        Whether this annotation is discoverable.

        :param discoverable: The discoverable of this AnnotationInputV1.
        :type: bool
        """

        self._discoverable = discoverable

    @property
    def document(self):
        """
        Gets the document of this AnnotationInputV1.
        This annotation's document.

        :return: The document of this AnnotationInputV1.
        :rtype: str
        """
        return self._document

    @document.setter
    def document(self, document):
        """
        Sets the document of this AnnotationInputV1.
        This annotation's document.

        :param document: The document of this AnnotationInputV1.
        :type: str
        """

        self._document = document

    @property
    def interests(self):
        """
        Gets the interests of this AnnotationInputV1.
        A list of IDs representing the annotation's items of interest

        :return: The interests of this AnnotationInputV1.
        :rtype: list[AnnotationInterestInputV1]
        """
        return self._interests

    @interests.setter
    def interests(self, interests):
        """
        Sets the interests of this AnnotationInputV1.
        A list of IDs representing the annotation's items of interest

        :param interests: The interests of this AnnotationInputV1.
        :type: list[AnnotationInterestInputV1]
        """

        self._interests = interests

    @property
    def name(self):
        """
        Gets the name of this AnnotationInputV1.
        Human readable name. Null or whitespace names are not permitted.

        :return: The name of this AnnotationInputV1.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this AnnotationInputV1.
        Human readable name. Null or whitespace names are not permitted.

        :param name: The name of this AnnotationInputV1.
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")

        self._name = name

    @property
    def replies_to(self):
        """
        Gets the replies_to of this AnnotationInputV1.
        The ID of the Annotation that this annotation replies to.

        :return: The replies_to of this AnnotationInputV1.
        :rtype: str
        """
        return self._replies_to

    @replies_to.setter
    def replies_to(self, replies_to):
        """
        Sets the replies_to of this AnnotationInputV1.
        The ID of the Annotation that this annotation replies to.

        :param replies_to: The replies_to of this AnnotationInputV1.
        :type: str
        """

        self._replies_to = replies_to

    @property
    def type(self):
        """
        Gets the type of this AnnotationInputV1.
        This annotation's type: Report or Journal. Defaults to Journal.

        :return: The type of this AnnotationInputV1.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Sets the type of this AnnotationInputV1.
        This annotation's type: Report or Journal. Defaults to Journal.

        :param type: The type of this AnnotationInputV1.
        :type: str
        """

        self._type = type

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, AnnotationInputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import queue
import threading

import pandas as pd
import numpy as np

from seeq.sdk43 import *

from . import _common
from . import _login


def pull(items, *, start=None, end=None, grid='15min', header='__auto__', group_by=None, capsules_as='auto',
         tz_convert=None, calculation=None, bounding_values=False, return_status_df=False, errors='raise', quiet=False):
    """
    Retrieves signal, condition or scalar data from Seeq Server and return it in a DataFrame.
    :param items: A DataFrame or Series containing ID and Type columns that can be used to identify the items to pull.
    This is usually created via a call to spy.search().
    :type items: pd.DataFrame, pd.Series
    :param start: The starting time for which to pull data. Note that Seeq will potentially return one additional
    row that is earlier than this time (if it exists), as a "bounding value" for interpolation purposes. This
    argument must be a string that pd.to_datetime() can parse, or a pd.Timestamp. If not provided, 'start' will
    default to 'end' minus 1 hour.
    :type start: str, pd.Timestamp
    :param end: The end time for which to pull data. Note that Seeq will potentially return one additional
    row that is earlier than this time (if it exists), as a "bounding value" for interpolation purposes. This
    argument must be a string that pd.to_datetime() can accept. If not provided, 'end' will default to now.
    :type end: str, pd.Timestamp
    :param grid: A period to use for interpolation such that all returned samples have the same timestamps.
    Interpolation will be applied at the server to achieve this. If grid=None is specified, no interpolation will occur
    and each signal's samples will be returned untouched. Where timestamps don't match, NaNs will be present within a
    row.
    :type grid: str, None
    :param header: The metadata property to use as the header of each column. Common values would be 'ID' or 'Name".
    Defaults to '__auto__', which concatenates Path and Name if they are present.
    :type header: str
    :param group_by: The name of a column or list of columns for which to group by. Often necessary when pulling data
    across assets: When you want header='Name', you typically need group_by=['Path', 'Asset']
    :type group_by: str, list(str)
    :param capsules_as: If 'auto' (default), returns capsules as a time series of 0 or 1 when signals are also
    present in the items argument, or returns capsules as individual rows if no signals are present. 'signals' or
    'capsules' forces the output to the former or the latter, if possible.
    :type capsules_as: str
    :param tz_convert: The time zone in which to return all timestamps. If the time zone string is not recognized,
    the list of supported time zone strings will be returned in the exception text.
    :type tz_convert: str
    :param calculation: When applying a calculation across assets, the 'calculation' argument must be a one-row
    DataFrame (or a Series) and the 'items' argument must be full of assets. When applying a calculation to a
    signal/condition/scalar, calculation must be a string with a signal variable in it: $signal, $condition or $scalar.
    :type calculation: str, pd.DataFrame
    :param bounding_values: If True, extra 'bounding values' will be returned before/after the specified query range
    for the purposes of assisting with interpolation to the edges of the range or, in the case of Step or PILinear
    interpolation methods, interpolating to 'now' when appropriate.
    :type bounding_values: bool
    :param return_status_df: If True, the return value will be a (data_df, status_df) tuple where data_df is the
    normal DataFrame of results and status_df is a copy of the items parameter with extra information about the pull
    operations, including a 'Result' column with error messages when errors='catalog'.
    :type return_status_df: bool
    :param errors: If 'raise' (default), any errors encountered will cause an exception. If 'catalog', errors will be
    added to a 'Result' column in the returned status_df DataFrame (errors='catalog' must be combined with
    return_status_df=True).
    :type errors: str
    :param quiet: If True, suppresses progress output.
    :type quiet: bool
    :return: A DataFrame with the requested data.
    :rtype: pd.DataFrame
    """
    _common.validate_timezone_arg(tz_convert)
    _common.validate_errors_arg(errors)

    if errors == 'catalog' and not return_status_df:
        raise RuntimeError("errors='catalog' must be combined with return_status_df=True")

    if not (isinstance(items, (pd.DataFrame, pd.Series))):
        raise RuntimeError('items argument must be a pandas.DataFrame or pandas.Series')

    if not items.index.is_unique:
        raise RuntimeError("The items DataFrame's index must be unique. Use reset_index(drop=True, inplace=True) "
                           "before passing in to spy.pull().")

    if isinstance(calculation, pd.DataFrame):
        if len(calculation) != 1:
            raise RuntimeError("When applying a calculation across assets, calculation argument must be a one-row "
                               "DataFrame, or a Series. When applying a calculation to a signal/condition/scalar, "
                               'calculation must be a string with a signal variable in it: $signal, $condition or '
                               '$scalar.')

        calculation = calculation.iloc[0]

    if isinstance(items, pd.Series):
        items = pd.DataFrame().append(items)

    if capsules_as not in ['auto', 'capsules', 'signals']:
        raise RuntimeError("capsules_as must be one of 'auto', 'capsules', 'signals'")

    # noinspection PyUnresolvedReferences
    pd_start = pd.to_datetime(start)  # type: pd.Timestamp
    # noinspection PyUnresolvedReferences
    pd_end = pd.to_datetime(end)  # type: pd.Timestamp

    if pd_end is None:
        pd_end = pd.to_datetime(pd.datetime.now())
        if pd_start is not None and pd_start > pd_end:
            pd_end = pd_start + pd.Timedelta(hours=1)

    if pd_start is None:
        pd_start = pd.to_datetime(pd.datetime.now()) if pd_end is None else pd_end
        pd_start = pd_start - pd.Timedelta(hours=1)

    status_columns = list()
    if 'ID' in items:
        status_columns.append('ID')
    if 'Path' in items:
        status_columns.append('Path')
    if 'Asset' in items:
        status_columns.append('Asset')
    if 'Name' in items:
        status_columns.append('Name')

    status_df = items[status_columns].copy()
    status_df['Count'] = 0
    status_df['Time'] = 0
    status_df['Result'] = 'Pulling'

    _common.display_status('Pulling data from %s to %s' % (pd_start, pd_end), _common.STATUS_RUNNING, quiet,
                           status_df)

    items_api = ItemsApi(_login.client)
    formulas_api = FormulasApi(_login.client)

    query_df = items  # type: pd.DataFrame
    output_df = pd.DataFrame()
    at_least_one_signal = False
    column_names = dict()
    final_column_names = list()
    for phase in ['signals', 'conditions', 'scalars', 'final']:
        threads = list()
        try:
            pulled_series = queue.Queue()
            status_updates = queue.Queue()

            def _drain_status_updates():
                while True:
                    try:
                        _index, _message, _exception, _count, _time = status_updates.get_nowait()

                        if _exception and errors == 'raise':
                            raise _exception

                        status_df.at[_index, 'Result'] = _message
                        status_df.at[_index, 'Count'] = _count
                        status_df.at[_index, 'Time'] = _time
                    except queue.Empty:
                        break

                _common.display_status(
                    'Pulling from <strong>%s</strong> to <strong>%s</strong>' % (pd_start, pd_end),
                    _common.STATUS_RUNNING, quiet, status_df)

            for index, row in query_df.iterrows():
                # noinspection PyBroadException
                try:
                    output_df, at_least_one_signal = _process_query_row(
                        _drain_status_updates, at_least_one_signal, calculation, capsules_as,
                        column_names, final_column_names, formulas_api, grid, header, index,
                        items_api, output_df, pd_start, pd_end, phase, pulled_series, row,
                        status_df, status_updates, threads, tz_convert, bounding_values)

                except BaseException:
                    if errors == 'raise':
                        raise

                    status_df.at[index, 'Result'] = _common.format_exception()

            still_running = True
            while still_running:
                _drain_status_updates()

                still_running = False
                for _, thread in threads:
                    thread.join(0.1)

                    if thread.is_alive():
                        still_running = True
                        break

            _drain_status_updates()

            # noinspection PyUnresolvedReferences
            for index, item_name, series in list(pulled_series.queue):
                item_row = query_df.loc[index]
                join_df = pd.DataFrame({item_name: series})
                if group_by is None:
                    if item_name in output_df.columns:
                        raise RuntimeError('Column headers not unique. 2+ instances of "%s" found. Use header="ID" '
                                           'to guarantee uniqueness, or alternatively try group_by=["Path", "Asset"] '
                                           'if you are using an asset tree.' % item_name)
                else:
                    if isinstance(group_by, str):
                        group_by = [group_by]
                    if not isinstance(group_by, list):
                        raise RuntimeError('groupby argument must be a str or list(str)')
                    if not all(col in query_df.columns for col in group_by):
                        raise RuntimeError('groupby columns %s not present in query DataFrame' % group_by)

                    for group_column in group_by:
                        join_df[group_column] = item_row[group_column]
                    join_df.set_index(group_by, inplace=True, append=True)

                output_df = join_df if len(output_df) == 0 else output_df.combine_first(join_df)

        except BaseException as e:
            for stop_event, _ in threads:
                stop_event.set()

            _common.status_exception(e, status_df, quiet)

            if errors == 'raise' and not isinstance(e, KeyboardInterrupt):
                raise

            return None

    _common.display_status('Pull successful from <strong>%s</strong> to <strong>%s</strong>' % (pd_start, pd_end),
                           _common.STATUS_SUCCESS, quiet, status_df)

    # Ensures that the order of the columns matches the order in the metadata
    output_df = output_df[final_column_names]

    return (output_df, status_df) if return_status_df else output_df


def _process_query_row(_drain_status_updates, at_least_one_signal, calculation, capsules_as, column_names,
                       final_column_names, formulas_api, grid, header, index, items_api, output_df, pd_start, pd_end,
                       phase, pulled_series, row, status_df, status_updates, threads, tz_convert, bounding_values):
    if phase == 'signals' and not _common.present(row, 'ID'):
        status_df.at[index, 'Result'] = 'No "ID" column - skipping'
        return output_df, at_least_one_signal

    item_id, item_name, item_type = _get_item_details(header, items_api, row)
    calculation_to_use = calculation
    if item_type == 'Asset':
        if calculation is None:
            raise RuntimeError('To pull data for an asset, you must provide a "calculate" argument whose '
                               'value is the metadata of a calculation that is based on a single asset.')

        swap_input = SwapInputV1()
        swap_input.swap_in = item_id
        calc_item_id, _, item_type = _get_item_details(header, items_api, calculation)

        item_dependency_output = items_api.get_formula_dependencies(
            id=calc_item_id)  # type: ItemDependencyOutputV1

        unique_assets = set(dep.ancestors[-1].id
                            for dep in item_dependency_output.dependencies
                            if len(dep.ancestors) > 0)

        if len(unique_assets) != 1:
            raise RuntimeError('To pull data for an asset, the "calculate" parameter must be a calculated '
                               'item that involves only one asset.')

        swap_input.swap_out = unique_assets.pop()

        swapped_item = items_api.find_swap(id=calc_item_id, body=[swap_input])  # type: ItemPreviewV1

        item_id = swapped_item.id

        # Don't try to apply a calculation later, we've already done it via our swap activity
        calculation_to_use = None

    if phase == 'signals' and \
            'Signal' not in item_type and 'Condition' not in item_type and 'Scalar' not in item_type:
        status_df.at[index, 'Result'] = 'Not a Signal, Condition or Scalar - skipping'
        return output_df, at_least_one_signal

    if 'Signal' in item_type:
        at_least_one_signal = True

    if phase == 'signals' and 'Signal' in item_type:

        parameters = ['signal=%s' % item_id]
        if calculation_to_use is not None:
            formula = calculation_to_use
        else:
            formula = '$signal'

        if grid:
            formula = 'resample(%s, %s)' % (formula, grid)

        stop_event = threading.Event()
        thread = threading.Thread(target=_pull_signal, args=(formulas_api, formula, parameters, item_id,
                                                             item_name, pulled_series, pd_start, pd_end,
                                                             tz_convert, column_names, stop_event,
                                                             status_updates, index, bounding_values))

        thread.start()
        threads.append((stop_event, thread))

    elif phase == 'conditions' and 'Condition' in item_type:
        if capsules_as == 'capsules' and at_least_one_signal:
            raise RuntimeError('Pull cannot include signals when conditions present and "capsules_as" '
                               'parameter is "capsules"')

        if capsules_as == 'auto':
            capsules_as = 'signals' if at_least_one_signal else 'capsules'

        if capsules_as == 'signals' and not at_least_one_signal:
            if grid is None:
                raise RuntimeError(
                    'Pull cannot include conditions when no signals present with capsules_as='
                    'capsules and grid=None')

            placeholder_item_name = '__placeholder__'
            _pull_signal(formulas_api, '0.toSignal(%s)' % grid, list(), placeholder_item_name,
                         placeholder_item_name, pulled_series, pd_start, pd_end, tz_convert, dict(),
                         threading.Event())

            _, _, series = pulled_series.get(True, 30)
            output_df[placeholder_item_name] = series

        parameters = ['condition=%s' % item_id]
        if calculation_to_use is not None:
            formula = calculation_to_use
        else:
            formula = '$condition'

        output_df = _pull_condition(capsules_as, formulas_api, formula, parameters, item_id,
                                    item_name, output_df, pulled_series, pd_start, pd_end, tz_convert,
                                    column_names, index, status_df, _drain_status_updates)

    elif phase == 'scalars' and 'Scalar' in item_type:
        parameters = ['scalar=%s' % item_id]
        if calculation_to_use is not None:
            formula = calculation_to_use
        else:
            formula = '$scalar'

        _pull_scalar(formulas_api, formula, parameters, item_id, item_name, output_df, column_names, index,
                     status_df, _drain_status_updates)

    elif phase == 'final':
        if item_id in column_names:
            for column_name in column_names[item_id]:
                if column_name not in final_column_names:
                    final_column_names.append(column_name)

    return output_df, at_least_one_signal


def _convert_column_timezone(ts_column, tz):
    ts_column = ts_column.tz_localize('UTC')
    return ts_column.tz_convert(tz) if tz else ts_column


def _pull_condition(capsules_as, formulas_api, formula, parameters, item_id, item_name, output_df, pulled_series,
                    pd_start, pd_end, tz, column_names, index, status_df, _drain_status_updates):
    # noinspection PyBroadException
    try:
        timer = _common.timer_start()
        capsule_count = 0
        current_start = pd_start.value
        offset = 0
        while True:
            formula_run_output, _, http_headers = formulas_api.run_formula_with_http_info(
                formula=formula,
                parameters=parameters,
                start='%d ns' % current_start,
                end='%d ns' % pd_end.value,
                offset=offset,
                limit=_common.DEFAULT_PULL_PAGE_SIZE)  # type: FormulaRunOutputV1

            next_start = current_start
            capsules_output = formula_run_output.capsules  # type: CapsulesOutputV1
            check_for_dupes = True

            columns = dict()
            if capsules_as == 'signals':
                columns[item_name] = pd.Series(0, output_df.index)
                for capsule in capsules_output.capsules:
                    pd_capsule_start = _common.convert_to_timestamp(
                        capsule.start if capsule.start is not None else 0, tz)
                    pd_capsule_end = _common.convert_to_timestamp(
                        capsule.end if capsule.end is not None else 7258118400000000000, tz)
                    columns[item_name].loc[(columns[item_name].index >= pd_capsule_start) &
                                           (columns[item_name].index <= pd_capsule_end)] = 1

                    for prop in capsule.properties:  # type: ScalarPropertyV1
                        colname = '%s - %s' % (item_name, prop.name)
                        if colname not in columns:
                            columns[colname] = pd.Series(np.nan, output_df.index)
                        columns[colname].loc[(columns[colname].index >= pd_capsule_start) &
                                             (columns[colname].index <= pd_capsule_end)] = prop.value

                column_names[item_id] = list()
                for col, series in columns.items():
                    pulled_series.put((index, col, series))
                    column_names[item_id].append(col)
            else:
                capsule_df_rows = list()

                for capsule in capsules_output.capsules:
                    column_names[item_id] = ['Condition', 'Capsule Start', 'Capsule End']
                    if check_for_dupes and \
                            'Condition' in output_df and \
                            'Capsule Start' in output_df and \
                            'Capsule End' in output_df and \
                            len(output_df.loc[(output_df['Condition'] == item_name) &
                                              (output_df['Capsule Start'] == _common.convert_to_timestamp(capsule.start,
                                                                                                          tz)) &
                                              (output_df['Capsule End'] == _common.convert_to_timestamp(capsule.end,
                                                                                                        tz))]):
                        # This can happen as a result of pagination
                        continue

                    check_for_dupes = False

                    capsule_dict = {
                        'Condition': item_name,
                        'Capsule Start': _common.convert_to_timestamp(capsule.start, tz),
                        'Capsule End': _common.convert_to_timestamp(capsule.end, tz)
                    }

                    for prop in capsule.properties:  # type: ScalarPropertyV1
                        capsule_dict[prop.name] = prop.value
                        column_names[item_id].append(prop.name)

                    capsule_df_rows.append(capsule_dict)

                    if not pd.isna(capsule.start) and capsule.start > next_start:
                        next_start = capsule.start

                output_df = output_df.append(capsule_df_rows)

            # Note that capsule_count here can diverge from the exact count in the output due to pagination
            capsule_count += len(capsules_output.capsules)

            if len(capsules_output.capsules) < _common.DEFAULT_PULL_PAGE_SIZE:
                break

            if next_start == current_start:
                # This can happen if the page is full of capsule that all have the same start time
                offset += _common.DEFAULT_PULL_PAGE_SIZE
            else:
                offset = 0

            current_start = next_start

            status_df.at[index, 'Result'] = 'Pulling %s' % _common.convert_to_timestamp(current_start, tz)
            status_df.at[index, 'Count'] = capsule_count
            status_df.at[index, 'Time'] = _common.timer_elapsed(timer)

            _drain_status_updates()

        status_df.at[index, 'Result'] = 'Success'
        status_df.at[index, 'Count'] = capsule_count
        status_df.at[index, 'Time'] = _common.timer_elapsed(timer)

    except BaseException:
        status_df.at[index, 'Result'] = _common.format_exception()

    return output_df


def _pull_signal(formulas_api, formula, parameters, item_id, item_name, pulled_series, pd_start, pd_end, tz,
                 column_names, stop_event=None, status_updates=None, index=None, bounding_values=False):
    # noinspection PyBroadException
    try:
        series = pd.Series()
        timer = _common.timer_start()
        current_start = pd_start
        last_key = 0
        while not stop_event.is_set():
            formula_run_output, _, http_headers = formulas_api.run_formula_with_http_info(
                formula=formula,
                parameters=parameters,
                start='%d ns' % current_start.value,
                end='%d ns' % pd_end.value,
                offset=0,
                limit=_common.DEFAULT_PULL_PAGE_SIZE)  # type: FormulaRunOutputV1

            series_samples_output = formula_run_output.samples  # type: SeriesSamplesOutputV1

            def _keep_sample(_sample_output):
                if _sample_output.key <= last_key:
                    return False

                if bounding_values:
                    return True

                if _sample_output.key < pd_start.value:
                    return False

                if _sample_output.key > pd_end.value:
                    return False

                return True

            time_index = _convert_column_timezone(pd.DatetimeIndex([sample_output.key for sample_output in
                                                                    series_samples_output.samples if
                                                                    _keep_sample(sample_output)]), tz)

            series = series.append(pd.Series([sample_output.value for sample_output in
                                              series_samples_output.samples if _keep_sample(sample_output)],
                                             index=time_index))

            if len(series_samples_output.samples) < _common.DEFAULT_PULL_PAGE_SIZE:
                break

            if len(series) > 0:
                last_key = series.index[-1].value

            if time_index[-1].value > current_start.value:
                current_start = time_index[-1]

            if status_updates is not None:
                status_updates.put((index, 'Pulling: %s' % str(current_start), None,
                                    len(series), _common.timer_elapsed(timer)))

        column_names[item_id] = [item_name]
        if status_updates is not None:
            status_updates.put((index, 'Success', None, len(series), _common.timer_elapsed(timer)))
        pulled_series.put((index, item_name, series))

    except BaseException as e:
        if status_updates is not None:
            status_updates.put((index, _common.format_exception(), e, None, None))


def _pull_scalar(formulas_api, formula, parameters, item_id, item_name, output_df, column_names, status_index,
                 status_df, _drain_status_updates):
    timer = _common.timer_start()

    formula_run_output, _, http_headers = formulas_api.run_formula_with_http_info(
        formula=formula,
        parameters=parameters)  # type: FormulaRunOutputV1

    if len(output_df.index) == 0:
        output_df.at[0, item_name] = formula_run_output.scalar.value
    else:
        output_df[item_name] = formula_run_output.scalar.value

    column_names[item_id] = [item_name]

    status_df.at[status_index, 'Result'] = 'Success'
    status_df.at[status_index, 'Count'] = 1
    status_df.at[status_index, 'Time'] = _common.timer_elapsed(timer)


def _get_item_details(header, items_api, row):
    item_id = _common.get(row, 'ID')
    item = None

    if _common.present(row, 'Type'):
        item_type = _common.get(row, 'Type')
    else:
        item = items_api.get_item_and_all_properties(id=item_id)  # type: ItemOutputV1
        item_type = item.type

    if header.upper() == 'ID':
        item_name = item_id
    elif _common.present(row, header):
        item_name = _common.get(row, header)
    else:
        if not item:
            item = items_api.get_item_and_all_properties(id=item_id)  # type: ItemOutputV1

        item_name = ''
        if header == '__auto__' and _common.present(row, 'Path'):
            item_name = _common.get(row, 'Path') + ' >> '
            if _common.present(row, 'Asset'):
                item_name += _common.get(row, 'Asset') + ' >> '

        if header in ['__auto__', 'Name']:
            item_name += item.name
        elif header == 'Description':
            item_name += item.description
        else:
            prop = [p.value for p in item.properties if p.name == header]
            if len(prop) == 0:
                item_name += item_id
            else:
                item_name += prop[0]

    return item_id, item_name, item_type



import logging
import os
import six
import urllib
import urllib3

from seeq.base import gconfig
from seeq.sdk43 import *
from seeq.sdk43.rest import ApiException

from . import _common

from urllib3.connectionpool import MaxRetryError

client = None  # type: ApiClient
user = None  # type: UserOutputV1

AUTOMATIC_PROXY_DETECTION = '__auto__'


def login(username=None, password=None, *, url=None, directory='Seeq', ignore_ssl_errors=False,
          proxy=AUTOMATIC_PROXY_DETECTION, credentials_file=None, auth_token=None, quiet=False, auth_provider=None):
    """
    Establishes a connection with Seeq Server and logs in with a set of credentials.
    :param username: Username for login purposes. See credentials_file argument for alternative.
    :type username: str
    :param password: Password for login purposes. See credentials_file argument for alternative.
    :type password: str
    :param url: Seeq Server url. You can copy this from your browser and cut off everything to the right of the port
    (if present). E.g. https://myseeqserver:34216
    :type url: str
    :param directory: The authentication directory to use. You must be able to supply a username/password, so some
    passwordless Windows Authentication (NTLM) scenarios will not work. OpenID Connect is also not supported. If you
    need to use such authentication schemes, set up a Seeq Data Lab server.
    :type directory: str
    :param ignore_ssl_errors: If True, SSL certificate validation errors are ignored. Use this if you're in a trusted
    network environment but Seeq Server's SSL certificate is not from a common root authority.
    :type ignore_ssl_errors: bool
    :param proxy: Specifies the proxy server to use for all requests. The default value is "__auto__", which examines
    the standard HTTP_PROXY and HTTPS_PROXY environment variables. If you specify None for this parameter,
    no proxy server will be used.
    :type proxy: str
    :param credentials_file: Reads username and password from the specified file. If specified,
    :type credentials_file: str
    :param auth_token: Provide an authorization token directly from a browser session of Seeq Workbench.
    :type auth_token: str
    :param quiet: If True, suppresses progress output.
    :type quiet: bool
    :param auth_provider: Deprecated. Use directory instead.
    """

    if auth_provider is not None:
        raise RuntimeError('"auth_provider" argument has been renamed to "directory"')

    # Annoying warnings are printed to stderr if connections fail
    logging.getLogger("requests").setLevel(logging.FATAL)
    logging.getLogger("urllib3").setLevel(logging.FATAL)
    urllib3.disable_warnings()

    if url:
        parsed_url = urllib.parse.urlparse(url)
        gconfig.override_global_property('seeq_server_hostname', parsed_url.hostname)
        if parsed_url.scheme == 'https':
            port = '443'
            if parsed_url.port:
                port = six.text_type(parsed_url.port)
            gconfig.override_global_property('seeq_server_port', '')
            gconfig.override_global_property('seeq_secure_port', port)
        if parsed_url.scheme == 'http':
            port = '80'
            if parsed_url.port:
                port = six.text_type(parsed_url.port)
            gconfig.override_global_property('seeq_server_port', port)
            gconfig.override_global_property('seeq_secure_port', '')

    api_client_url = gconfig.get_api_url()

    cert_file = os.path.join(gconfig.get_data_folder(), 'keys', 'seeq-cert.pem')
    if os.path.exists(cert_file):
        Configuration().cert_file = cert_file
    key_file = os.path.join(gconfig.get_data_folder(), 'keys', 'seeq-key.pem')
    if os.path.exists(key_file):
        Configuration().key_file = key_file

    Configuration().verify_ssl = not ignore_ssl_errors

    if proxy == AUTOMATIC_PROXY_DETECTION:
        if api_client_url.startswith('https') and 'HTTPS_PROXY' in os.environ:
            Configuration().proxy = os.environ['HTTPS_PROXY']
        elif 'HTTP_PROXY' in os.environ:
            Configuration().proxy = os.environ['HTTP_PROXY']
    elif proxy is not None:
        Configuration().proxy = proxy

    global client
    client = ApiClient(api_client_url)

    if auth_token:
        if username or password or credentials_file:
            raise RuntimeError('username, password and/or credentials_file cannot be provided along with auth_token')

        client.auth_token = auth_token
    else:
        auth_api = AuthApi(client)
        auth_input = AuthInputV1()

        if credentials_file:
            if username is not None or password is not None:
                raise RuntimeError('If credentials_file is specified, username and password must be None')

            try:
                with open(credentials_file) as f:
                    lines = f.readlines()
            except Exception as e:
                raise RuntimeError('Could not read credentials_file "%s": %s' % (credentials_file, e))

            if len(lines) < 2:
                raise RuntimeError('credentials_file "%s" must have two lines: username then password')

            username = lines[0].strip()
            password = lines[1].strip()

        if not username or not password:
            raise RuntimeError('Both username and password must be supplied')

        auth_input.username = username
        auth_input.password = password

        _common.display_status('Logging in to <strong>%s</strong> as <strong>%s</strong>' % (
            api_client_url, username), _common.STATUS_RUNNING, quiet)

        directories = dict()
        try:
            auth_providers_output = auth_api.get_auth_providers()  # type: AuthProvidersOutputV1
        except MaxRetryError as e:
            raise RuntimeError(
                '"%s" could not be reached. Is the server or network down?\n%s' % (api_client_url, e))

        for datasource_output in auth_providers_output.auth_providers:  # type: DatasourceOutputV1
            directories[datasource_output.name] = datasource_output

        if directory not in directories:
            raise RuntimeError('directory "%s" not recognized. Possible directory(s) for this server: %s' %
                               (directory, ', '.join(directories.keys())))

        datasource_output = directories[directory]
        auth_input.auth_provider_class = datasource_output.datasource_class
        auth_input.auth_provider_id = datasource_output.datasource_id

        try:
            auth_api.login(body=auth_input)
        except ApiException as e:
            if e.status == 401:
                raise RuntimeError(
                    '"%s" could not be logged in with supplied credentials, check username and password.' %
                    auth_input.username)
            else:
                raise
        except MaxRetryError as e:
            raise RuntimeError(
                '"%s" could not be reached. Is the server or network down?\n%s' % (api_client_url, e))
        except Exception as e:
            raise RuntimeError('Could not connect to Seeq"s API at %s with login "%s".\n%s' % (api_client_url,
                                                                                               auth_input.username, e))

    users_api = UsersApi(client)

    global user
    user = users_api.get_me()  # type: UserOutputV1

    user_string = user.username
    user_profile = ''
    if user.first_name:
        user_profile = user.first_name
    if user.last_name:
        user_profile += ' ' + user.last_name
    if user.is_admin:
        user_profile += ' [Admin]'
    if len(user_profile) > 0:
        user_string += ' (%s)' % user_profile.strip()

    _common.display_status('Logged in to <strong>%s</strong> successfully as <strong>%s</strong>' % (
        api_client_url, user_string), _common.STATUS_SUCCESS, quiet)


def logout(quiet=False):
    """
    Logs you out of your current session.
    :param quiet: If True, suppresses progress output.
    :type quiet: bool
    """
    global client  # type: ApiClient
    if client is None:
        _common.display_status('No action taken because you are not currently logged in.', _common.STATUS_FAILURE,
                               quiet)

    auth_api = AuthApi(client)
    auth_api.logout()

    client.logout()

    _common.display_status('Logged out.', _common.STATUS_SUCCESS, quiet)



from .. import _common
from ._workbook import *
from ... import spy


def pull(workbooks_df, *, include_referenced_workbooks=True, return_status_df=False, errors='raise', quiet=False):
    _common.validate_errors_arg(errors)

    for required_column in ['ID', 'Type', 'Workbook Type']:
        if required_column not in workbooks_df.columns:
            raise RuntimeError('"%s" column must be included in workbooks_df' % required_column)

    status_columns = list()

    for col in ['ID', 'Name', 'Workbook Type']:
        if col in workbooks_df:
            status_columns.append(col)

    workbooks_df = workbooks_df[workbooks_df['Type'] != 'Folder']

    status_df = workbooks_df[status_columns].copy().reset_index(drop=True)
    status_df['Count'] = 0
    status_df['Time'] = 0
    status_df['Result'] = 'Queued'
    status_columns.extend(['Count', 'Time', 'Result'])

    _common.display_status('Pulling workbooks', _common.STATUS_RUNNING, quiet, status_df)

    results = list()
    workbooks_to_pull = dict()
    referencing_search_folder_id = dict()
    original_item_map = dict()
    for index, row in workbooks_df.iterrows():
        item_id = _common.get(row, 'ID')

        try:
            if _common.get(row, 'Workbook Type') == 'Analysis':
                if item_id not in workbooks_to_pull:
                    workbooks_to_pull[item_id] = set()

                continue

            timer = _common.timer_start()
            status_df.at[status_df['ID'] == item_id, 'Result'] = 'Pulling'

            workbook = Item.pull(item_id,
                                 original_item_map=original_item_map,
                                 allowed_types=['Workbook'],
                                 status=(status_df, timer, quiet))  # type: Workbook

            if include_referenced_workbooks and _common.get(row, 'Workbook Type') == 'Topic':
                for workbook_id, workstep_tuple in workbook.referenced_workbooks.items():
                    if workbook_id not in workbooks_to_pull:
                        workbooks_to_pull[workbook_id] = set()
                        if len(status_df[status_df['ID'] == workbook_id]) == 0:
                            search_df = spy.workbooks.search({'ID': workbook_id}, quiet=True)
                            if len(search_df) == 1:
                                search_df['Count'] = 0
                                search_df['Time'] = 0
                                search_df['Result'] = 'Queued'
                                status_df.loc[len(status_df)] = search_df.iloc[0][status_columns]

                    if workstep_tuple not in workbooks_to_pull[workbook_id]:
                        workbooks_to_pull[workbook_id].update(workstep_tuple)
                        referencing_search_folder_id[workbook_id] = _common.get(row, 'Search Folder ID')

            if _common.present(row, 'Search Folder ID'):
                workbook['Search Folder ID'] = _common.get(row, 'Search Folder ID')

            results.append(workbook)

            status_df.at[status_df['ID'] == item_id, 'Time'] = _common.timer_elapsed(timer)
            status_df.at[status_df['ID'] == item_id, 'Result'] = 'Success'

        except BaseException as e:
            if isinstance(e, KeyboardInterrupt):
                status_df['Result'] = 'Canceled'
                status_message = 'Pull canceled'
                status_code = _common.STATUS_CANCELED
                _common.display_status(status_message, status_code, quiet, status_df)
                return None

            if errors == 'raise':
                raise

            status_df.at[status_df['ID'] == item_id, 'Result'] = _common.format_exception(e)

    for workbook_id, workstep_tuples in workbooks_to_pull.items():
        timer = _common.timer_start()

        try:
            workbook = Item.pull(workbook_id,
                                 original_item_map=original_item_map,
                                 allowed_types=['Workbook'],
                                 status=(status_df, timer, quiet))  # type: Workbook

            workbook.pull_worksteps(list(workstep_tuples))

            workbook_row = workbooks_df[workbooks_df['ID'] == workbook_id]
            if len(workbook_row) == 1 and 'Search Folder ID' in workbook_row.columns:
                workbook['Search Folder ID'] = workbook_row.iloc[0]['Search Folder ID']
            elif workbook_id in referencing_search_folder_id and referencing_search_folder_id[workbook_id] is not None:
                # If the workbook was pulled only as a result of being referenced by something else (like a Topic),
                # we don't have a specific Search Folder ID to use. So just use the search folder ID of the referencing
                # item. (Note: If it's referenced by multiple things, then "last one wins".)
                workbook['Search Folder ID'] = referencing_search_folder_id[workbook_id]

            results.append(workbook)

            status_df.at[status_df['ID'] == workbook_id, 'Time'] = _common.timer_elapsed(timer)
            status_df.at[status_df['ID'] == workbook_id, 'Result'] = 'Success'

        except BaseException as e:
            if isinstance(e, KeyboardInterrupt):
                status_df['Result'] = 'Canceled'
                status_message = 'Pull canceled'
                status_code = _common.STATUS_CANCELED
                _common.display_status(status_message, status_code, quiet, status_df)
                return None

            if errors == 'raise':
                raise

            status_df.at[status_df['ID'] == workbook_id, 'Result'] = _common.format_exception(e)

    for workbook in results:
        workbook.use_original_ids(original_item_map)

    _common.display_status('Pull successful', _common.STATUS_SUCCESS, quiet, status_df)

    return (results, status_df) if return_status_df else results



"""
Short for Seeq PYthon, the Spy library provides methods to interact with data that is exposed to the Seeq Server.
"""

from . import assets
from . import docs
from . import workbooks

from ._login import login
from ._login import logout
from ._plot import plot
from ._pull import pull
from ._push import push
from ._search import search
from ._common import DEFAULT_WORKBOOK_PATH

__all__ = ['assets', 'docs', 'workbooks', 'login', 'logout', 'plot', 'pull', 'push', 'search', 'DEFAULT_WORKBOOK_PATH']



import pytest
import sys

from seeq import spy
from seeq.spy.assets import Asset, Mixin

from . import test_common


def setup_module():
    test_common.login()


class HVAC(Asset):

    @Asset.Attribute()
    def Temperature(self, metadata):
        # We use simple Pandas syntax to select for a row in the DataFrame corresponding to our desired tag
        return metadata[metadata['Name'].str.endswith('Temperature')]

    @Asset.Attribute()
    def Relative_Humidity(self, metadata):
        # All Attribute functions must take (self, metadata) as parameters
        return metadata[metadata['Name'].str.contains('Humidity')]


class Compressor(Asset):

    @Asset.Attribute()
    def Power(self, metadata):
        return metadata[metadata['Name'].str.endswith('Power')]


class Airflow_Attributes(Mixin):
    @Asset.Attribute()
    def Airflow_Rate(self, metadata):
        return {
            'Type': 'Signal',
            'Formula': 'sinusoid()'
        }


class HVAC_With_Calcs(HVAC):

    @Asset.Attribute()
    def Temperature_Rate_Of_Change(self, metadata):
        return {
            'Type': 'Signal',

            # This formula will give us a nice derivative in F/h
            'Formula': '$temp.lowPassFilter(150min, 3min, 333).derivative() * 3600 s/h',

            'Formula Parameters': {
                # We can reference the base class' Temperature attribute here as a dependency
                '$temp': self.Temperature(metadata),
            }
        }

    @Asset.Attribute()
    def Too_Hot(self, metadata):
        return {
            'Type': 'Condition',
            'Formula': '$temp.valueSearch(isGreaterThan($threshold))',
            'Formula Parameters': {
                '$temp': self.Temperature(metadata),

                # We can also reference other attributes in this derived class
                '$threshold': self.Hot_Threshold(metadata)
            }
        }

    @Asset.Attribute()
    def Hot_Threshold(self, metadata):
        return {
            'Type': 'Scalar',
            'Formula': '80F'
        }

    # Designating a function as a Component allows you to include a child asset with its own set of attributes
    @Asset.Component()
    def Compressor(self, metadata):
        return Compressor(parent=self)

    @Asset.Component()
    def Pump(self, metadata):
        return [
            {
                'Name': 'Pump Volume',
                'Type': 'Scalar',
                'Formula': '1000L'
            },
            {
                'Name': 'Pump Voltage',
                'Type': 'Scalar',
                'Formula': '110V'
            }
        ]

    @Asset.Component()
    def Airflow(self, metadata):
        return Airflow_Attributes(self)


@pytest.mark.system
def test_build():
    hvac_metadata_df = spy.search({
        'Name': 'Area ?_*',
        'Datasource Class': 'Time Series CSV Files'
    })

    hvac_metadata_df['Build Template'] = 'HVAC'

    hvac_metadata_df['Build Asset'] = hvac_metadata_df['Name'].str.extract('(Area .)_.*')

    hvac_metadata_df['Build Path'] = 'My HVAC Units >> Facility #1'

    # It won't like the "Build Template" column since we're specifying the HVAC class directly
    with pytest.raises(RuntimeError):
        spy.assets.build(HVAC, hvac_metadata_df)

    hvac_metadata_df = hvac_metadata_df.drop(columns=['Build Template'])
    build_df = spy.assets.build(HVAC, hvac_metadata_df)

    # We'll get an error the first time because Area F doesn't have the signals we need
    with pytest.raises(RuntimeError):
        spy.push(metadata=build_df)

    # Now we'll catalog the errors instead of stopping on them
    spy.push(metadata=build_df, errors='catalog')

    hvac_with_calcs_metadata_df = hvac_metadata_df.copy()

    build_with_calcs_df = spy.assets.build(HVAC_With_Calcs, hvac_with_calcs_metadata_df)

    push_results_df = spy.push(metadata=build_with_calcs_df, errors='catalog')

    errors_df = push_results_df[push_results_df['Push Result'] != 'Success']

    # Should only be 4 errors (associated with Area F)
    assert len(errors_df) == 4

    search_results_df = spy.search({
        'Path': 'My HVAC Units >> Facility #1'
    })

    areas = [
        'Area A',
        'Area B',
        'Area C',
        'Area D',
        'Area E',
        'Area F',
        'Area G',
        'Area H',
        'Area I',
        'Area J',
        'Area K',
        'Area Z',
    ]

    for area in areas:
        assertions = [
            ('My HVAC Units >> Facility #1', area, 'Temperature', 'CalculatedSignal'),
            ('My HVAC Units >> Facility #1', area, 'Temperature Rate Of Change', 'CalculatedSignal'),
            ('My HVAC Units >> Facility #1', area, 'Relative Humidity', 'CalculatedSignal'),
            ('My HVAC Units >> Facility #1', area, 'Too Hot', 'CalculatedCondition'),
            ('My HVAC Units >> Facility #1', area, 'Hot Threshold', 'CalculatedScalar'),
            ('My HVAC Units >> Facility #1', area, 'Pump Voltage', 'CalculatedScalar'),
            ('My HVAC Units >> Facility #1', area, 'Pump Volume', 'CalculatedScalar'),
            ('My HVAC Units >> Facility #1 >> ' + area, 'Compressor', 'Power', 'CalculatedSignal'),
            ('My HVAC Units >> Facility #1', area, 'Airflow Rate', 'CalculatedSignal'),
        ]

        # Area F is special!
        if area == 'Area F':
            assertions = [
                ('My HVAC Units >> Facility #1', area, 'Hot Threshold', 'CalculatedScalar'),
                ('My HVAC Units >> Facility #1', area, 'Pump Voltage', 'CalculatedScalar'),
                ('My HVAC Units >> Facility #1', area, 'Pump Volume', 'CalculatedScalar'),
                ('My HVAC Units >> Facility #1 >> ' + area, 'Compressor', 'Power', 'CalculatedSignal'),
            ]

        for _path, _asset, _name, _type in assertions:
            assertion_df = search_results_df[
                (search_results_df['Path'] == _path) &
                (search_results_df['Asset'] == _asset) &
                (search_results_df['Name'] == _name) &
                (search_results_df['Type'] == _type)]

            assert len(assertion_df) == 1, \
                'Instantiated item not found: %s, %s, %s, %s' % (_path, _asset, _name, _type)


@pytest.mark.system
def test_build_with_module():
    hvac_metadata_df = spy.search({
        'Name': 'Area ?_*',
        'Datasource Class': 'Time Series CSV Files'
    })

    hvac_metadata_df['Build Path'] = 'My HVAC Units >> Facility #2'

    def _template_chooser(name):
        if 'Compressor' in name:
            return 'Compressor'
        else:
            return 'HVAC'

    hvac_metadata_df['Build Template'] = hvac_metadata_df['Name'].apply(_template_chooser)

    hvac_metadata_df['Area'] = hvac_metadata_df['Name'].str.extract('(Area .)_.*')
    hvac_metadata_df['Build Asset'] = hvac_metadata_df['Area'] + ' ' + hvac_metadata_df['Build Template']

    build_df = spy.assets.build(sys.modules[__name__], hvac_metadata_df)

    spy.push(metadata=build_df)

    search_results_df = spy.search({
        'Path': 'My HVAC Units >> Facility #2'
    })

    # There should be a "Area X HVAC" and "Area X Compressor" signals
    assert len(search_results_df) == 57



import pytest

from seeq import spy

from ...tests import test_common


def setup_module():
    # test_common.login()
    pass


@pytest.mark.system2
def test_save_load():
    # spy.login(url='https://explore.seeq.com', username='mark.derbecker@seeq.com', password='RR!Harley',
    # ignore_ssl_errors=True)

    spy.login(username='mark.derbecker@seeq.com', password='SeeQ2013!')

    search_df = spy.workbooks.search({
        'Path': 'Swap Folder'
    })

    #pull_df = spy.workbooks.pull(search_df)
    #spy.workbooks.save(pull_df, r'D:\Scratch\WorkbookExport2', clean=True)

    pull_df = spy.workbooks.load(r'D:\Scratch\WorkbookExport2')
    status_df = spy.workbooks.push(pull_df, use_full_path=True, to_original_owner=True, replace_acl=True)



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: 0.44.00-BETA
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat
from six import iteritems
import re


class SystemConfigurationOutputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'default_auth_auto_login': 'bool',
        'default_auth_provider_id': 'str',
        'default_max_capsule_duration': 'ScalarValueOutputV1',
        'default_max_interpolation': 'ScalarValueOutputV1',
        'maximum_threads': 'int',
        'number_format': 'str',
        'parallelization_threshold': 'ScalarValueOutputV1',
        'priorities': 'list[PriorityV1]',
        'registration_enabled': 'bool',
        'restrict_logs': 'bool',
        'status_message': 'str'
    }

    attribute_map = {
        'default_auth_auto_login': 'defaultAuthAutoLogin',
        'default_auth_provider_id': 'defaultAuthProviderId',
        'default_max_capsule_duration': 'defaultMaxCapsuleDuration',
        'default_max_interpolation': 'defaultMaxInterpolation',
        'maximum_threads': 'maximumThreads',
        'number_format': 'numberFormat',
        'parallelization_threshold': 'parallelizationThreshold',
        'priorities': 'priorities',
        'registration_enabled': 'registrationEnabled',
        'restrict_logs': 'restrictLogs',
        'status_message': 'statusMessage'
    }

    def __init__(self, default_auth_auto_login=False, default_auth_provider_id=None, default_max_capsule_duration=None, default_max_interpolation=None, maximum_threads=None, number_format=None, parallelization_threshold=None, priorities=None, registration_enabled=False, restrict_logs=False, status_message=None):
        """
        SystemConfigurationOutputV1 - a model defined in Swagger
        """

        self._default_auth_auto_login = None
        self._default_auth_provider_id = None
        self._default_max_capsule_duration = None
        self._default_max_interpolation = None
        self._maximum_threads = None
        self._number_format = None
        self._parallelization_threshold = None
        self._priorities = None
        self._registration_enabled = None
        self._restrict_logs = None
        self._status_message = None

        if default_auth_auto_login is not None:
          self.default_auth_auto_login = default_auth_auto_login
        if default_auth_provider_id is not None:
          self.default_auth_provider_id = default_auth_provider_id
        if default_max_capsule_duration is not None:
          self.default_max_capsule_duration = default_max_capsule_duration
        if default_max_interpolation is not None:
          self.default_max_interpolation = default_max_interpolation
        if maximum_threads is not None:
          self.maximum_threads = maximum_threads
        if number_format is not None:
          self.number_format = number_format
        if parallelization_threshold is not None:
          self.parallelization_threshold = parallelization_threshold
        if priorities is not None:
          self.priorities = priorities
        if registration_enabled is not None:
          self.registration_enabled = registration_enabled
        if restrict_logs is not None:
          self.restrict_logs = restrict_logs
        if status_message is not None:
          self.status_message = status_message

    @property
    def default_auth_auto_login(self):
        """
        Gets the default_auth_auto_login of this SystemConfigurationOutputV1.
        True if the system should automatically attempt to login without user action. Only applies when the default auth provider datasource class is \"Windows Auth\"

        :return: The default_auth_auto_login of this SystemConfigurationOutputV1.
        :rtype: bool
        """
        return self._default_auth_auto_login

    @default_auth_auto_login.setter
    def default_auth_auto_login(self, default_auth_auto_login):
        """
        Sets the default_auth_auto_login of this SystemConfigurationOutputV1.
        True if the system should automatically attempt to login without user action. Only applies when the default auth provider datasource class is \"Windows Auth\"

        :param default_auth_auto_login: The default_auth_auto_login of this SystemConfigurationOutputV1.
        :type: bool
        """

        self._default_auth_auto_login = default_auth_auto_login

    @property
    def default_auth_provider_id(self):
        """
        Gets the default_auth_provider_id of this SystemConfigurationOutputV1.
        The datasource ID of the authentication provider that should be automatically selected when the Seeq user interface is displayed

        :return: The default_auth_provider_id of this SystemConfigurationOutputV1.
        :rtype: str
        """
        return self._default_auth_provider_id

    @default_auth_provider_id.setter
    def default_auth_provider_id(self, default_auth_provider_id):
        """
        Sets the default_auth_provider_id of this SystemConfigurationOutputV1.
        The datasource ID of the authentication provider that should be automatically selected when the Seeq user interface is displayed

        :param default_auth_provider_id: The default_auth_provider_id of this SystemConfigurationOutputV1.
        :type: str
        """

        self._default_auth_provider_id = default_auth_provider_id

    @property
    def default_max_capsule_duration(self):
        """
        Gets the default_max_capsule_duration of this SystemConfigurationOutputV1.
        The scalar value of the default capsule duration

        :return: The default_max_capsule_duration of this SystemConfigurationOutputV1.
        :rtype: ScalarValueOutputV1
        """
        return self._default_max_capsule_duration

    @default_max_capsule_duration.setter
    def default_max_capsule_duration(self, default_max_capsule_duration):
        """
        Sets the default_max_capsule_duration of this SystemConfigurationOutputV1.
        The scalar value of the default capsule duration

        :param default_max_capsule_duration: The default_max_capsule_duration of this SystemConfigurationOutputV1.
        :type: ScalarValueOutputV1
        """

        self._default_max_capsule_duration = default_max_capsule_duration

    @property
    def default_max_interpolation(self):
        """
        Gets the default_max_interpolation of this SystemConfigurationOutputV1.
        The scalar value of the default maximum interpolation

        :return: The default_max_interpolation of this SystemConfigurationOutputV1.
        :rtype: ScalarValueOutputV1
        """
        return self._default_max_interpolation

    @default_max_interpolation.setter
    def default_max_interpolation(self, default_max_interpolation):
        """
        Sets the default_max_interpolation of this SystemConfigurationOutputV1.
        The scalar value of the default maximum interpolation

        :param default_max_interpolation: The default_max_interpolation of this SystemConfigurationOutputV1.
        :type: ScalarValueOutputV1
        """

        self._default_max_interpolation = default_max_interpolation

    @property
    def maximum_threads(self):
        """
        Gets the maximum_threads of this SystemConfigurationOutputV1.
        The maximum number of threads to use when parallelizing a single request

        :return: The maximum_threads of this SystemConfigurationOutputV1.
        :rtype: int
        """
        return self._maximum_threads

    @maximum_threads.setter
    def maximum_threads(self, maximum_threads):
        """
        Sets the maximum_threads of this SystemConfigurationOutputV1.
        The maximum number of threads to use when parallelizing a single request

        :param maximum_threads: The maximum_threads of this SystemConfigurationOutputV1.
        :type: int
        """

        self._maximum_threads = maximum_threads

    @property
    def number_format(self):
        """
        Gets the number_format of this SystemConfigurationOutputV1.
        The default number format to be used for items without a number format property.

        :return: The number_format of this SystemConfigurationOutputV1.
        :rtype: str
        """
        return self._number_format

    @number_format.setter
    def number_format(self, number_format):
        """
        Sets the number_format of this SystemConfigurationOutputV1.
        The default number format to be used for items without a number format property.

        :param number_format: The number_format of this SystemConfigurationOutputV1.
        :type: str
        """

        self._number_format = number_format

    @property
    def parallelization_threshold(self):
        """
        Gets the parallelization_threshold of this SystemConfigurationOutputV1.
        The duration, in nanoseconds, above which requests may be parallelized

        :return: The parallelization_threshold of this SystemConfigurationOutputV1.
        :rtype: ScalarValueOutputV1
        """
        return self._parallelization_threshold

    @parallelization_threshold.setter
    def parallelization_threshold(self, parallelization_threshold):
        """
        Sets the parallelization_threshold of this SystemConfigurationOutputV1.
        The duration, in nanoseconds, above which requests may be parallelized

        :param parallelization_threshold: The parallelization_threshold of this SystemConfigurationOutputV1.
        :type: ScalarValueOutputV1
        """

        self._parallelization_threshold = parallelization_threshold

    @property
    def priorities(self):
        """
        Gets the priorities of this SystemConfigurationOutputV1.
        The priorities for metrics, sorted descending by level

        :return: The priorities of this SystemConfigurationOutputV1.
        :rtype: list[PriorityV1]
        """
        return self._priorities

    @priorities.setter
    def priorities(self, priorities):
        """
        Sets the priorities of this SystemConfigurationOutputV1.
        The priorities for metrics, sorted descending by level

        :param priorities: The priorities of this SystemConfigurationOutputV1.
        :type: list[PriorityV1]
        """

        self._priorities = priorities

    @property
    def registration_enabled(self):
        """
        Gets the registration_enabled of this SystemConfigurationOutputV1.
        True if new user registration is available from the Seeq login screen

        :return: The registration_enabled of this SystemConfigurationOutputV1.
        :rtype: bool
        """
        return self._registration_enabled

    @registration_enabled.setter
    def registration_enabled(self, registration_enabled):
        """
        Sets the registration_enabled of this SystemConfigurationOutputV1.
        True if new user registration is available from the Seeq login screen

        :param registration_enabled: The registration_enabled of this SystemConfigurationOutputV1.
        :type: bool
        """

        self._registration_enabled = registration_enabled

    @property
    def restrict_logs(self):
        """
        Gets the restrict_logs of this SystemConfigurationOutputV1.
        True if log access is restricted to admins.

        :return: The restrict_logs of this SystemConfigurationOutputV1.
        :rtype: bool
        """
        return self._restrict_logs

    @restrict_logs.setter
    def restrict_logs(self, restrict_logs):
        """
        Sets the restrict_logs of this SystemConfigurationOutputV1.
        True if log access is restricted to admins.

        :param restrict_logs: The restrict_logs of this SystemConfigurationOutputV1.
        :type: bool
        """

        self._restrict_logs = restrict_logs

    @property
    def status_message(self):
        """
        Gets the status_message of this SystemConfigurationOutputV1.

        :return: The status_message of this SystemConfigurationOutputV1.
        :rtype: str
        """
        return self._status_message

    @status_message.setter
    def status_message(self, status_message):
        """
        Sets the status_message of this SystemConfigurationOutputV1.

        :param status_message: The status_message of this SystemConfigurationOutputV1.
        :type: str
        """

        self._status_message = status_message

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, SystemConfigurationOutputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import pytest

import pandas as pd

from seeq import spy
from seeq.sdk43.rest import ApiException

from . import test_common

from .. import _common


def setup_module():
    test_common.login()


@pytest.mark.system
def test_simple_search():
    search_results = spy.search({
        'Name': 'Area A_Temper'
    })

    assert len(search_results) == 1

    search_results = spy.search(pd.DataFrame([{
        'Name': 'Area A_Temper'
    }]))

    # Nothing will be returned because we use a equal-to comparison when a DataFrame is passed in
    assert len(search_results) == 0

    search_results = spy.search(pd.DataFrame([{
        'Name': 'Area A_Temperature'
    }]))

    assert len(search_results) == 1


@pytest.mark.system
def test_dataframe_single_row_with_id():
    search_results = spy.search({
        'Name': 'Area A_Temper'
    })

    search_results = spy.search(search_results.iloc[0])

    assert len(search_results) == 1
    assert search_results.iloc[0]['Name'] == 'Area A_Temperature'
    assert search_results.iloc[0]['Data ID'] == '[Tag] Area A_Temperature.sim.ts.csv'


@pytest.mark.system
def test_dataframe_multi_row():
    search_results = spy.search(pd.DataFrame([{
        'Name': 'Area A_Temperature',
        'Datasource Name': 'Example Data'
    }, {
        'Path': 'Example >> Cooling Tower 1 >> Area A',
        'Name': 'Relative Humidity'
    }]))

    assert len(search_results) == 2
    assert len(search_results[search_results['Name'] == 'Area A_Temperature']) == 1
    assert len(search_results[search_results['Name'] == 'Relative Humidity']) == 1


@pytest.mark.system
def test_path_with_datasource():
    search_results = spy.search({
        'Path': 'Example >> Cooling Tower 1 >> Area A',
        'Name': 'Relative Humidity',
        'Datasource Name': 'Example Data'
    })

    assert len(search_results) == 1
    assert len(search_results[search_results['Name'] == 'Relative Humidity']) == 1

    search_results = spy.search({
        'Path': 'Example >> Cooling Tower 1 >> Area A',
        'Name': 'Relative Humidity',
        'Datasource Name': 'Seeq Database',
        'Datasource Class': 'cassandraV2'
    })

    assert len(search_results) == 0


@pytest.mark.system
def test_dataframe_bad_datasource():
    with pytest.raises(RuntimeError):
        spy.search(pd.DataFrame([{
            'Name': 'Area A_Temperature',
            'Datasource Name': 'Bad Datasource'
        }]))


@pytest.mark.system
def test_type_search():
    search_results = spy.search({
        'Datasource Class': 'Time Series CSV Files',
        'Type': 'Signal'
    })

    assert 150 < len(search_results) < 200

    datasource_names = set(search_results['Datasource Name'].tolist())
    assert len(datasource_names) == 1
    assert datasource_names.pop() == 'Example Data'

    types = set(search_results['Type'].tolist())
    assert len(types) == 1
    assert types.pop() == 'StoredSignal'

    search_results = spy.search({
        'Datasource Class': 'Time Series CSV Files',
        'Type': 'Condition'
    })

    assert len(search_results) == 0

    search_results = spy.search({
        'Datasource Class': 'Time Series CSV Files',
        'Type': 'Scalar'
    })

    assert len(search_results) == 0

    search_results = spy.search({
        'Datasource Class': 'Time Series CSV Files',
        'Type': 'Asset'
    })

    assert 5 < len(search_results) < 20

    # Multiple types
    search_results = spy.search({
        'Datasource Class': 'Time Series CSV Files',
        'Type': ['Signal', 'Asset']
    })

    assert 160 < len(search_results) < 300
    assert 5 < len(search_results[search_results['Type'] == 'Asset']) < 20
    assert 150 < len(search_results[search_results['Type'].str.contains('Signal')]) < 200


@pytest.mark.system
def test_path_search_recursive():
    with pytest.raises(RuntimeError):
        spy.search({
            'Path': 'Non-existent >> Path'
        })

    search_results = spy.search({
        'Path': 'Example >> Cooling Tower 1'
    })

    assert 40 < len(search_results) < 60


@pytest.mark.system
def test_path_search_non_recursive():
    with pytest.raises(RuntimeError):
        spy.search({
            'Name': 'Blah',
            'Path': 'Example >> Cooling Tower 1'
        }, recursive=False)

    search_results = spy.search({
        'Path': 'Example >> Cooling Tower 1'
    }, recursive=False, workbook=None)

    assert 5 < len(search_results) < 10


@pytest.mark.system
def test_path_search_pagination():
    # This tests the 'Path' finding code to make sure we'll find a path even if pagination
    # is required.
    original_page_size = _common.DEFAULT_SEARCH_PAGE_SIZE
    try:
        _common.DEFAULT_SEARCH_PAGE_SIZE = 1
        search_results = spy.search({
            'Path': 'Example >> Cooling Tower 1 >> Area G'
        }, recursive=False, workbook=None)

        assert len(search_results) == 6
    finally:
        _common.DEFAULT_SEARCH_PAGE_SIZE = original_page_size


@pytest.mark.system
def test_path_search_bad_path():
    try:
        spy.search({
            'Path': 'Exclample >> Cooling Tower Bogus'
        }, recursive=False)

        assert False, 'Expected RuntimeError exception'

    except RuntimeError as e:
        error_string = str(e)
        assert 'Could not find item "Exclample" as root asset' in error_string
        assert '"Example"' in error_string

    try:
        spy.search({
            'Path': 'Example >> Cooling Tower Bogus'
        }, recursive=False)

        assert False, 'Expected RuntimeError exception'

    except RuntimeError as e:
        assert 'Could not find item "Cooling Tower Bogus" under "Example"' in str(e)
        assert '"Cooling Tower 1"' in str(e)
        assert '"Cooling Tower 2"' in str(e)


@pytest.mark.system
def test_datasource_name_search():
    with pytest.raises(RuntimeError):
        spy.search({
            'Datasource Name': 'Non-existent'
        })

    search_results = spy.search({
        'Datasource Name': 'Example Data'
    })

    assert 150 < len(search_results) < 200


@pytest.mark.system
def test_search_pagination():
    original_page_size = _common.DEFAULT_SEARCH_PAGE_SIZE
    try:
        _common.DEFAULT_SEARCH_PAGE_SIZE = 2
        search_results = spy.search({
            'Name': 'Area A_*'
        })

        assert len(search_results) == 6

    finally:
        _common.DEFAULT_SEARCH_PAGE_SIZE = original_page_size


@pytest.mark.system
def test_search_bad_workbook():
    with pytest.raises(RuntimeError):
        spy.search({
            'Name': 'Area A_*'
        }, workbook='bad')


@pytest.mark.system
def test_search_workbook_guid():
    # The workbook won't be found, so we'll get an access error
    with pytest.raises(ApiException):
        spy.search({
            'Name': 'Area A_*'
        }, workbook='A0B89103-E95D-4E32-A809-390C1FAE9D2F')



import pytest

from seeq import spy

from ...tests import test_common


def setup_module():
    test_common.login()


@pytest.mark.system4
def test_search():
    spy.login('mark.derbecker@seeq.com', 'SeeQ2013!')
    results_df = spy.workbooks.search(path='Swap Folder', all_properties=True, recursive=True)
    print(results_df)



import os

from .. import _common
from ._workbook import *


def load(folder):
    if not os.path.exists(folder):
        raise RuntimeError('Folder "%s" does not exist' % folder)

    workbook_json_files = glob.glob(os.path.join(folder, '**', 'Workbook.json'), recursive=True)

    workbooks = list()
    for workbook_json_file in workbook_json_files:
        workbooks.append(Workbook.load(os.path.dirname(workbook_json_file)))

    return workbooks
